class RadialHomographyMapping implements PlanarMapping
{
  PMatrix3D homography = new PMatrix3D(1.0, 0.1, 0.1, 0.0, 0.5, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0);
  float lambda = 0.00001;
  
  PVector f( PVector _p )
  {
    PVector p = new PVector();
    p.x = _p.x;
    p.y = _p.y;    
    p.z = 1 + lambda * (p.x * p.x + p.y * p.y);
    
    PVector q = new PVector();
    homography.mult(p, q);                 
    
    q.x /= q.z;
    q.y /= q.z;
    q.z  = 1.0;
    
    return q;
  }
  
  PVector df( int i, int j, PVector p )
  {
    float u = p.x; 
    float v = p.y;
    float x = u;
    float y = v;
    
    float h1 = homography.m00;
    float h2 = homography.m01;
    float h3 = homography.m02;
    float h4 = homography.m10;
    float h5 = homography.m11;
    float h6 = homography.m12;
    float h7 = homography.m20;
    float h8 = homography.m21;
    float h9 = homography.m22;
    
    float x2 = x*x;
    float x3 = x*x2;
    float x4 = x2*x2;
    float y2 = y*y;
    float y3 = y*y2;
    float y4 = y2*y2;
    float lambda2 = lambda*lambda;
    float lambda3 = lambda*lambda2;
    float lambda4 = lambda2*lambda2;
    float h12 = h1*h1;
    float h22 = h2*h2;
    float h32 = h3*h3;
    float h42 = h4*h4;
    float h52 = h5*h5;
    float h62 = h6*h6;
    float h72 = h7*h7;
    float h82 = h8*h8;
    float h92 = h9*h9;
    
    float denom = (h9*(lambda*(y2+x2)+1)+h8*y+h7*x);
    float denom2 = denom*denom;
    float denom3 = denom*denom2;
    float denom4 = denom2*denom2;
    float factorA = (2*lambda*h9*x+h7);
    float factorA2 = factorA*factorA;
    float factorA3 = factorA*factorA2;
    float factorB = (2*lambda*h9*y+h8);
    float factorB2 = factorB*factorB;
    float factorB3 = factorB*factorB2;
    
    if ( i == 1 && j == 0 )
    {
      PVector result = new PVector( (2*lambda*h3*x+h1)/denom-((2*lambda*h9*x+h7)*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom2,
                                    (2*lambda*h6*x+h4)/denom-((2*lambda*h9*x+h7)*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom2 );
      result.x *= 100;
      result.y *= 100;
      return result;
    }

    if ( i == 0 && j == 1 )
    {
      PVector result = new PVector( (2*lambda*h3*y+h2)/denom-((2*lambda*h9*y+h8)*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom2,
                                    (2*lambda*h6*y+h5)/denom-((2*lambda*h9*y+h8)*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom2 );
      result.x *= 200;
      result.y *= 200;
      return result;
    }
    
    if ( i == 2 && j == 0 )
    {
      PVector result = new PVector( (2*lambda*h3)/denom-(2*lambda*h9*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom2-(2*(2*lambda*h3*x+h1)*factorA)/denom2+(2*factorA2*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom3,
                                    (2*lambda*h6)/denom-(2*lambda*h9*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom2-(2*(2*lambda*h6*x+h4)*factorA)/denom2+(2*factorA2*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom3 );
      result.x *= 200;
      result.y *= 200;
      return result;
    }
    
    if ( i == 0 && j == 2 )
    {
      PVector result = new PVector( (2*lambda*h3)/denom-(2*lambda*h9*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom2-(2*(2*lambda*h3*y+h2)*factorB)/denom2+(2*factorB2*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom3,
                                    (2*lambda*h6)/denom-(2*lambda*h9*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom2-(2*(2*lambda*h6*y+h5)*factorB)/denom2+(2*factorB2*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom3 );
      result.x *= 200;
      result.y *= 200;
      return result;
    }
    
    if ( i == 3 && j == 0 )
    {
      PVector result = new PVector( -(6*lambda*h3*factorA)/denom2-(6*lambda*h9*(2*lambda*h3*x+h1))/denom2+(12*lambda*h9*factorA*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom3+(6*(2*lambda*h3*x+h1)*factorA2)/denom3-(6*factorA3*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom4,
                                    -(6*lambda*h6*factorA)/denom2-(6*lambda*h9*(2*lambda*h6*x+h4))/denom2+(12*lambda*h9*factorA*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom3+(6*(2*lambda*h6*x+h4)*factorA2)/denom3-(6*factorA3*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom4 );
      result.x *= 200;
      result.y *= 200;
      return result;
    }
    
    if ( i == 0 && j == 3 )
    {
      PVector result = new PVector( -(6*lambda*h3*factorB)/denom2-(6*lambda*h9*(2*lambda*h3*y+h2))/denom2+(6*(2*lambda*h3*y+h2)*factorB2)/denom3+(12*lambda*h9*factorB*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom3-(6*factorB3*(h3*(lambda*(y2+x2)+1)+h2*y+h1*x))/denom4,
                                    -(6*lambda*h6*factorB)/denom2-(6*lambda*h9*(2*lambda*h6*y+h5))/denom2+(6*(2*lambda*h6*y+h5)*factorB2)/denom3+(12*lambda*h9*factorB*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom3-(6*factorB3*(h6*(lambda*(y2+x2)+1)+h5*y+h4*x))/denom4 );
      result.x *= 200;
      result.y *= 200;
      return result;
    }      
    
    // all mixed partials are zero
    return new PVector( 0, 0 );
  }
  
  PVector boundDf( int n )
  {
    float m = 0;
    return new PVector( m, m );
  }
}
